% BERGEN & HEEGER 2D TEXTURE SYSTHESIS
% Heeger & Bergen (1995). Pyramid Based Texture Analysis/Synthesis., Computer Graphics Proceedings. p. 229-238.
%
% You can enter color or grayscale images.
%
% The final synthetic texture is in the matrix 'textureColor'

% You need to install the steerable pyramide toolbox first.
% http://www.cns.nyu.edu/~eero/steerpyr/

% Antonio Torralba, 2004 Email: torralba@csail.mit.edu

clear all


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FILE NAME OF THE TRAINING TEXTURE:
f='192026.jpg';

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% PARAMETERS:
Nbins=256; % number of bins for computing histograms
Niter=15; % number of iterations
pyrFilters = 'sp3Filters'; Nor=4; Nscales=4;
%pyrFilters = 'sp5Filters'; Nor=6; Nscales=3; 
%pyrFilters = 'sp1Filters'; Nor=1; Nscales=3;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Read image:
exampleTraining=imread(f);
[nx,ny,c] = size(exampleTraining); 
n = min([nx ny]);
m = n; % lets make the new texture to have the same size than the target (we can change this)

figure
exampleTraining=double(exampleTraining(1:n,1:n,:));
imagesc(exampleTraining/256); colormap(gray(256)); axis('equal'); axis('tight')
drawnow

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ANALYSIS: COLLECT STATISTICS OF TARGET TEXTURE:
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 1) Substract mean value to each color channel and store the value:
color = size(exampleTraining,3);
for i=1:color
    MeanColor(i)=mean(mean(exampleTraining(:,:,i))); % we will need to put this back later
    example(:,:,i)=exampleTraining(:,:,i)-MeanColor(i);
end

% 2) PCA of color space to have decorrelated color channels.
if color==3
    % If the color channels were independent, then we could treat each
    % channel as a different texture. We are going to assume that this is
    % true. In reality it is not true but it works well. You could create
    % color images in which this assumption will produce really bad
    % results. 
    clear x
    for i=1:color
        ex=example(:,:,i);
        x(:,i)=ex(:);
    end
    X=x'*x;
    [pc,latent] = eigs(X,3);
    %x=(pc'*x')';
    x = x*pc;
    for i=1:color
        example(:,:,i)=reshape(x(:,i),[n n]);
    end
end


% 3) COLLECT SUBBAND HISTOGRAMS OF TARGET TEXTURE:
for i=1:color
    exampleC=example(:,:,i);
    [examplePyr, exampleIndT] = buildSpyr(exampleC, Nscales, pyrFilters, 'reflect1'); 
    % Attention to the boundary conditions (we reflect the border pixels for the analysis).
    figure
    showSpyr(examplePyr, exampleIndT); 
    title('Pyramid decomposition of training texture')

    % Precompute subband histograms:
    % 1) Histogram of the image:
    minIma(i)=min(exampleC(:));
    maxIma(i)=max(exampleC(:));
    [H,bins]=hist((exampleC(:)-minIma(i))/(maxIma(i)-minIma(i)),256);
    HistIma(:,i)=H';
    
    % 2) Histogram of the subbands (There are Nscales * Nor subands and two residuals):
    for b=1:Nscales*Nor+2
        subband =  pyrBand(examplePyr,exampleIndT,b); 
        subband = subband(3:end-2,3:end-2); % Remove some artifacts at the image boundary

        % Puts intensity values in the interval [0,1] (more or less) and
        % compute the histogram:
        s = std(subband(:));
        minb(b,i)=-4*s;
        maxb(b,i)= 4*s; 
        [H,bins]=hist((subband(:)-minb(b,i))/(maxb(b,i)-minb(b,i)), linspace(0,1,Nbins));
        Histb(:,b,i)=H';
    end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TEXTURE SYNTESIS (with color):
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INITIAL TEXTURE ( you can change this initialization to be anything you
% want). It could be another texture.
textureC=250*rand(m,m,color); 

scrsz = get(0,'ScreenSize');
fig1=figure('name','Syntetic texture', 'Position', [10 scrsz(4)/2 fix(scrsz(3)/2.2) fix(scrsz(4)/2.4)]);
fig2=figure('name','Subband histograms (red = target histogram, blue = current iteration)', 'Position', [10+fix(scrsz(3)/2.2) 10 fix(scrsz(3)/2.2) fix(scrsz(4)/2.4)]);
fig3=figure('name','Evolution of one subband during the first six iterations', 'Position', [10 10 fix(scrsz(3)/2.2) fix(scrsz(4)/2.4)]);
fig4=figure('name','Evolution of the synthetic texture during the first six iterations', 'Position', [10+fix(scrsz(3)/2.2) scrsz(4)/2 fix(scrsz(3)/2.2) fix(scrsz(4)/2.4)]);

for iter=1:Niter
    for i=1:color
        texture=textureC(:,:,i);
        
        % Coerce image (for each color channel) histogram:
        texture=(texture-min(texture(:)))/(max(texture(:))-min(texture(:)));
        texture=histeq(texture(:),HistIma(:,i)/max(HistIma(:,i)));
        texture=reshape(texture*(maxIma(i)-minIma(i))+minIma(i),[m m]);

        % Coerce subband statistics:
        [texturePyr, textureIndT] = buildSpyr(texture, Nscales, pyrFilters, 'circular'); % Now we use circular convolution instead of reflecting the boundary.
        figure(fig2)
        if i == 1
            subplot(Nscales+2, Nor, 1);
            imagesc(texture); colormap(gray(256)); axis('equal'); axis('tight'); axis('off')
        end
        for b=1:Nscales*Nor+2                       
            % Coerce subband histogram:
            x =  pyrBand(texturePyr,textureIndT,b);

            [mb,nb]=size(x);
            
            s = std(x(:)); Mn = 4*s; mn = -4*s;
            x=(x-mn)/(Mn-mn); % puts values (more or less) in the interval [0,1]
            
            if i==1
                % only plot one channel so that there are not too many figures...
                subplot(Nscales+2, Nor, b+Nor-1);
                h = hist(x(:), linspace(0,1,Nbins)); h = h/sum(h);
                plot(h); hold on; plot(Histb(:,b,i)/ sum(Histb(:,b,i)),'r'); axis('tight'); hold off
                if b == 1; title('High pass residual'); end
                if b == Nscales*Nor+2; xlabel('Low pass residual'); end
                drawnow
            end
            
            x=histeq(x(:),Histb(:,b,i)/max(Histb(:,b,i)));
            x=reshape(x*(maxb(b,i)-minb(b,i))+minb(b,i),[mb nb]);
            
             if i==1
                if b == 3
                    if iter <7
                        figure(fig3)
                        
                        subplot(2, 3, iter);
                        imagesc(x); colormap(gray(256)); axis('equal'); axis('tight'); axis('off')
                        figure(fig2)
                    end
                    
                end
            end
            
            indices = pyrBandIndices(textureIndT, b);
            texturePyr(indices)=x(:);
            

        end
        texture = reconSpyr(texturePyr, textureIndT, pyrFilters, 'circular');
        
        
        % Put texture in the corresponding color channel:
        textureC(:,:,i)=texture;
    end
    
    % Reconstruct color channels from PCA:
    if color==3
        clear x
        for i=1:color
            ex=textureC(:,:,i);
            x(:,i)=ex(:);
        end
        x=x*pc'; % Rotate back to RGB color space
        textureColor=zeros(size(textureC));
        for i=1:color
            textureColor(:,:,i)=reshape(x(:,i),[n n])+MeanColor(i); % put back the mean
        end
    else 
        textureColor=texture+MeanColor(1); 
    end
    
    % Visualization of the texture at each iteration:
    mn=min([textureColor(:); exampleTraining(:)]);
    Mx=max([textureColor(:); exampleTraining(:)]);
    
    figure(fig1)
    subplot(121)
    image(uint8(255*(exampleTraining-mn)/(Mx-mn)))
    axis('square')
    xlabel('Original texture')
    subplot(122)

    image(uint8(255*(textureColor-mn)/(Mx-mn)))
    title(iter)
    xlabel('Synthetic texture')
    axis('square')
    colormap(gray(256))
    drawnow
    
    if iter<7
        figure(fig4)
        subplot(2, 3, iter);
        image(uint8(255*(textureColor-mn)/(Mx-mn)))
        title(iter)
        axis('square')
        colormap(gray(256))
    end
end



